var gc = gc || {};

(function() {

    var Category = function(name, list) {
        this.name = name;
        this.list = list;
    };

    Category.prototype.addElement = function() {
        this.list.push.apply(this.list, arguments);
        this.list.sort(sortElementsfn);
    };

    Category.prototype.getElements = function() {
        return this.list;
    };

    Category.prototype.clone = function(elementsToAdd) {
        var result = new Category(this.name, this.list.slice());
        result.list.push.apply(result.list, elementsToAdd);
        result.list.sort(sortElementsfn);
        return result;
    };

    var resolveCategory = function(categories, element) {
        if (!element.isHidden) {
            var categoryName = element.group;

            if (!categoryName) {
                categoryName = 'Other Components';
            } else {
                categoryName = categoryName.split('/').pop();
            }

            if (!categories.hasOwnProperty(categoryName)) {
                categories[categoryName] = new Category(categoryName, []);
            }
            categories[categoryName].addElement(element)
        }
    };

    var ElementDatabase = function() {
        this.clear();
    };

    ElementDatabase.prototype.addElement = function(element) {
        resolveCategory(this.categoryMap, element);

        if (element.id) {
            this.elementMap[element.id] = element;
        }
    };

    ElementDatabase.prototype.byId = function(id) {
        return this.elementMap[id];
    };

    ElementDatabase.prototype.getCategories = function() {
        return this.categoryMap;
    };

    ElementDatabase.prototype.getElementMap = function() {
        return this.elementMap;
    };

    ElementDatabase.prototype.clear = function() {
        this.categoryMap = {};
        this.elementMap = {};
    };

    var metadata_v2 = {
        standardElements: new ElementDatabase(),
        projectMap: {}
    };
    var metadata_v3 = {
        standardElements: new ElementDatabase(),
        projectMap: {}
    };
    var metadata_dyn = {
        standardElements: new ElementDatabase(),
        projectMap: {},
        clear: function() {
            this.standardElements.clear();
            this.projectMap = {};
        }
    };

    var WORKSPACE_PATH = '/workspace/';

    var sortElementsfn = function(a, b) {
        var aa = (a.label || a.id).toLowerCase(), bb = (b.label || b.id).toLowerCase();
        return aa < bb ? -1 : (aa > bb ? 1 : 0);
    };

    var sortCategoriesfn = function(a, b) {
        var aa = a.name.toLowerCase(), bb = b.name.toLowerCase();
        if (aa < bb) {
            return aa === 'other components' ? 1 : -1;
        } else if (aa > bb) {
            return bb === 'other components' ? -1 : 1;
        }
        return 0;
    };

    var toCamelCaseRegEx = /-([a-z])/g;

    var toUpperCase = function(m) {
        return m[1].toUpperCase();
    };

    var convertAttributeToProperty = function(attributeName) {
        return attributeName.replace(toCamelCaseRegEx, toUpperCase);
    };

    var copyAttributes = function(p, convertFn) {
        var o = {};
        convertFn = convertFn || function(arg) { return arg; };

        Array.prototype.forEach.call(p.attributes, function(a) {
            var v = a.value;
            v = v.indexOf(',') > 0 ? v.split(',') : v;
            // handle boolean attribute
            if (v === '' || v === 'true') {
                v = true;
            } else if (v === 'false') {
                v = false;
            }
            o[convertFn(a.name)] = v;
        });
        return o;
    };

    var copyInheritedProperties = function(inheritedMeta, props, id, inherits) {
        if (inheritedMeta && inheritedMeta.metaProperties) {
            var inheritedProps = inheritedMeta.metaProperties;
            for(var propName in inheritedProps) {
                if (inheritedProps.hasOwnProperty(propName)) {
                    props[propName] = inheritedProps[propName];
                }
            }
        } else {
            console.error('x-meta: missing metadata for ' + inherits + ', for component ' + id);
        }
    };

    var queryProperties = function(node, props) {
        var ps = node.querySelectorAll('property');
        ps && Array.prototype.forEach.call(ps, function(p) {
            var o = copyAttributes(p);
            props[o.name] = o;
        });
    };

    var fitlerCategories = function(category, filterByText) {
        if (category.name.toLowerCase().indexOf(filterByText) >= 0) {
            return category;
        } else {
            var filteredList = [];
            var elements = category.getElements();
            for (var i = 0; i < elements.length; i++) {
                var element = elements[i];
                var lcLabel = element.label.toLowerCase();
                if (lcLabel.indexOf(filterByText) > -1) {
                    filteredList.push(element);
                }
            }
            if (filteredList.length > 0) {
                return new Category(category.name, filteredList);
            }
        }
    };

    var getMetaData = function (version) {
        switch (version) {
            case 3: return metadata_v3;
            case 2: return metadata_v2;
            default: return metadata_dyn;
        }
    }


    var XMetaImpl = {
        register: function(node) {
            var projectUrl;
            var elements = getMetaData(node.version).standardElements;
            var element = copyAttributes(node, convertAttributeToProperty);
            element.importPath = node._getImportPath();

            var pos = element.importPath.indexOf(WORKSPACE_PATH);
            if (pos >= 0) {
                // handle custom elements
                pos = pos + WORKSPACE_PATH.length;
                pos = element.importPath.indexOf('/', pos+1);
                pos = element.importPath.indexOf('/', pos+1);
                pos = element.importPath.indexOf('/', pos+1);
                projectUrl = element.importPath.substring(0, pos+1);

                if (!getMetaData(node.version).projectMap.hasOwnProperty(projectUrl)) {
                    getMetaData(node.version).projectMap[projectUrl] = new ElementDatabase();
                }
                elements = getMetaData(node.version).projectMap[projectUrl];
            } else {
                pos = element.importPath.indexOf('components/');
                if (pos > 0) {
                    element.importPath = element.importPath.substring(pos);
                }
            }

            element.metaProperties = this.getMetaProperties(node, projectUrl);
            element.importDependencies = this.getImportDependencies(node);

            var archetype = (typeof Polymer !== 'undefined' ? Polymer.dom(node) : node).querySelector('template');
            if (archetype) {
                element.archetype = archetype.innerHTML.trim();
                element.archetype = element.archetype.split('\r').join('');
            }

            elements.addElement(element);
            return element;
        },

        getMetaProperties: function(node, projectUrl) {
            var props = {};
            /* query inherited properties - must be executed first so that element properties can override the behavior properties */
            var bs = node.querySelectorAll('inherit');
            var that = this;
            bs && Array.prototype.forEach.call(bs, function(b) {
                copyInheritedProperties(that.byId(node, b.attributes.id.value, projectUrl), props, that.id, b.attributes.id.value);
            });

            /* query element properties and override any properties that already exist in the inherited properties */
            queryProperties(node, props);
            return props;
        },

        byId: function(node, id, projectBaseUrl) {
            if (id) {
                var result;
                if (projectBaseUrl && getMetaData(node.version).projectMap.hasOwnProperty(projectBaseUrl)) {
                    result = getMetaData(node.version).projectMap[projectBaseUrl].byId(id);
                }
                return result || getMetaData(node.version).standardElements.byId(id);
            }
        },

        getImportDependencies: function(node) {
            var template = typeof Polymer !== 'undefined' ?  Polymer.dom(node).querySelector('#imports') : node.getElementById('imports');
            var links = template && template.content && template.content.children || [];
            var result = [];
            for(var i = 0; i < links.length; i++ )
            {
                var link = links[i];
                if (link.localName === 'link')
                {
                    var href = link.getAttribute('href');
                    if (href && href !== '')
                    {
                        if (node.resolveRelativeUrl) {
                            href = node.resolveRelativeUrl(href);
                        } else {
                            href = resolveRelativeUrl(template.baseURI, href);
                        }
                        var pos = href.indexOf(WORKSPACE_PATH);
                        if (pos >= 0 )
                        {
                            pos = pos + WORKSPACE_PATH.length;
                            pos = href.indexOf('/', pos+1);  // need at least one character for user login name.
                            pos = href.indexOf('/', pos+1);  // need at least one character for user's project name.
                            pos = href.indexOf('/', pos+1);  // need at least one character form component version.
                            href = pos > 0 ? href.substring(pos+1) : href;
                        }
                        else if (href.indexOf('../../components/') === 0)
                        {
                            href = href.substring('../../'.length);
                        }
                        result.push(href);
                    }
                }
            }
            return result;
        },

        getCategories: function(node, projectBaseUrl, filterByText) {
            var result = [];
            var standardCategories = getMetaData(node.version).standardElements.getCategories();
            var customCategories = projectBaseUrl && getMetaData(node.version).projectMap.hasOwnProperty(projectBaseUrl) ? getMetaData(node.version).projectMap[projectBaseUrl].getCategories() : {};
            filterByText = filterByText ? filterByText.toLowerCase() : undefined;

            // add standard categories first, merging custom categories of the same name
            var categoryName;
            for(categoryName in standardCategories) {
                if (standardCategories.hasOwnProperty(categoryName)) {
                    var standardCategory = standardCategories[categoryName];
                    var customCategory = customCategories[categoryName];
                    if (customCategory) {
                        standardCategory = standardCategory.clone(customCategory.getElements());
                    }
                    if (filterByText) {
                        standardCategory = fitlerCategories(standardCategory, filterByText);
                    }
                    if (standardCategory) {
                        result.push(standardCategory);
                    }
                }
            }

            // add custom categories, that were not already merged with a standard category.
            for(categoryName in customCategories) {
                if (!standardCategories.hasOwnProperty(categoryName) && customCategories.hasOwnProperty(categoryName)) {
                    var category = customCategories[categoryName];
                    if (filterByText) {
                        category = fitlerCategories(category, filterByText);
                    }
                    if (category) {
                        result.push(category);
                    }
                }
            }

            result.sort(sortCategoriesfn);
            return result;
        },

        loadDynElements: function(elementMap) {
            metadata_dyn.clear();
            for(var elementName in elementMap) {
                if (elementMap.hasOwnProperty(elementName)) {
                    metadata_dyn.standardElements.addElement(elementMap[elementName]);
                }
            }
        },

        addElements: function(elementV2Map, elementV3Map) {
            for(var elementName in elementV2Map) {
                if (elementV2Map.hasOwnProperty(elementName)) {
                    metadata_v2.standardElements.addElement(elementV2Map[elementName]);
                }
            }
            if (elementV3Map) {
                for (var elementName in elementV3Map) {
                    if (elementV3Map.hasOwnProperty(elementName)) {
                        metadata_v3.standardElements.addElement(elementV3Map[elementName]);
                    }
                }
            }
        },

        addV3xElement: function(projUrl, elementMetadata) {
            var elements = getMetaData(3).projectMap[projUrl];
            if (!elements) {
                elements = getMetaData(3).projectMap[projUrl] = new ElementDatabase();
            }

            if (typeof elementMetadata.isHidden === 'undefined') elementMetadata.isHidden = false;
            if (typeof elementMetadata.isContainer === 'undefined') elementMetadata.isHidden = false;
            for (var e in elementMetadata.metaProperties) {
                var prop = elementMetadata.metaProperties[e];
                if (typeof prop.isHidden === 'undefined') {
                    prop.isHidden = false;
                }
            }

            elementMetadata.archetype = elementMetadata.archetype.trim().split('\r').join('');
            elements.addElement(elementMetadata);
        },

        getAllStandardElementsMap: function() {
            return getMetaData().standardElements.getElementMap();
        }
    };

    // The following were copied from Polymer 0.5 [...
    var baseUrl = function (node) {
        var u = new URL(node.ownerDocument.baseURI);
        u.search = '';
        u.hash = '';
        return u;
    };

    var resolveRelativeUrl = function (baseUrl, url, keepAbsolute) {
        // do not resolve '/' absolute urls
        if (url && url[0] === '/') {
            return url;
        }
        // do not resolve '#' links, they are used for routing
        if (url && url[0] === '#') {
            return url;
        }
        // do not resolve 'components', they are routed by the server
        if (url && url.indexOf('components') !== -1) {
            var index = url.indexOf('components');
            return url.substring(index);
        }
        var u = new URL(url, baseUrl);
        return keepAbsolute ? u.href : makeDocumentRelPath(u.href);
    };

    var makeDocumentRelPath = function (url) {
        var root = baseUrl(document.documentElement);
        var u = new URL(url, root);
        if (u.host === root.host && u.port === root.port &&
                u.protocol === root.protocol) {
            return makeRelPath(root, u);
        } else {
            return url;
        }
    };

    // make a relative path from source to target
    var makeRelPath = function (sourceUrl, targetUrl) {
        var source = sourceUrl.pathname;
        var target = targetUrl.pathname;
        var s = source.split('/');
        var t = target.split('/');
        while (s.length && s[0] === t[0]) {
            s.shift();
            t.shift();
        }
        for (var i = 0, l = s.length - 1; i < l; i++) {
            t.unshift('..');
        }
        // empty '#' is discarded but we need to preserve it.
        var URL_HASH = '#';
        var hash = (targetUrl.href.slice(-1) === URL_HASH) ? URL_HASH : targetUrl.hash;
        return t.join('/') + targetUrl.search + hash;
    };
//...]

    // support for require() node api.
    if (typeof module !== 'undefined') {
        module.exports = XMetaImpl;
    }
    else {
        gc['x-meta'] = XMetaImpl;
    }

}())
